/*
 *  create a list of character frequences from a specified file.
 *
 *  Arguments:
 *
 *    getcIn -- a pointer to a function returning a stream of chars
 *               that takes two arguments:
 *                        c -      address of next char to return from stream
 *                        status - address of a status flag
 *                     and returns TRUE iff accessing stream produced a char
 *
 *              the stream in question contains char frequency info,
 *              and is divided into lines consisting of two values:
 *                 1st in pair:  name of character, in octal (range: 1-277)
 *                 2nd in pair:  frequency count, a positive unsigned value
 *                    [initial frequency counts of zero disallowed, due to
 *                     assumptions in dynamic tree-balancing algorithm]
 *
 *              preconditions on input stream:  
 *                 stream should have unique count for each character
 *                      (precondition not audited by routine)
 *                 stream line length should not exceed 80 characters
 *                      (violation of condition flagged by routine)
 *
 *    status - a pointer to a boolean
 *
 *  Outputs:  
 *    if conversion fails, status is set to FALSE,
 *       and explanatory message is issued to stderr.
 *
 *    if conversion succeeds, status is set to TRUE,
 *        and routine returns a ptr to a list of <character,frequency> pairs
 *
 */

#include <stdio.h>            /* standard stream i/o package  */

#include "fltypes.h"

#include "fcfns.h"
#include "fifns.h"
#include "flfns.h"

/* change following define if longer input lines desired */
#define LINE_LENGTH 80

pFList flinit( 
  BOOLEAN (* getcIn) ( char *c, BOOLEAN *status ), BOOLEAN *status )
{ 
  pFList pfl = nilFList;        /* returned to caller:  the list per se     */
  unsigned linenum = 0;         /* track line number for error messages     */
  char linebuf[LINE_LENGTH+1];  /* store data from stream before conversion */
                                /* reserve one extra slot to add final null */

  BOOLEAN beforeEOF = TRUE;     /* exit condition: no more in stream        */

  /* loop:  get one more line from file */
  do {
     linenum++;      /* record step to next line  */

     /*
      *   get next line from stream, truncating line if needed  
      *   "-1" saves room for final null terminator
      */
     beforeEOF = getline( getcIn, linebuf, sizeof(linebuf)-1,
                          "frequency count", linenum);

     /*  skip empty lines:  could warn, but why bother?  */

     if (linebuf[0] != '\0') {
        unsigned charnum;        /* octal char codes from file stored here  */
        FCount freq;             /* frequency counts from file stored here  */
        int matchCount;          /* checking file's syntax, with fscanf() */

        /*  get counts data from line in linebuf  */
        matchCount=sscanf(linebuf,"%o %u\n",&charnum,&freq); 
 
        if (matchCount != 2) {
           fprintf(stderr,
                      "?? flinit: invalid file format at line %d\n",linenum);
           fprintf(stderr,"?? expected octal char num, then unsigned val\n");
           fldestroy(pfl);         /* deallocate allocated items */
           *status = FALSE;        /* routine failed             */
           return nilFList;
        }
 
        if (charnum > 255) {
           fprintf(stderr,"?? flinit: invalid file format\n");
           fprintf(stderr,
                      "?? character number %o not in range 0-277\n",charnum);
           fldestroy(pfl);         /* deallocate allocated items */
           *status = FALSE;        /* routine failed             */
           return nilFList;
        }
   
        if (freq == 0) {
           fprintf(stderr,"?? flinit: invalid frequency count data\n");
           fprintf(stderr,"?? initial frequency for %o is zero\n",charnum);
           fldestroy(pfl);         /* deallocate allocated items */
           *status = FALSE;        /* routine failed             */
           return nilFList;
        }
    
        {/* glue info about newly scanned item to head of list */
         pFList listElt = (pFList) malloc(sizeof(FList));
   
         putFLNext(listElt,pfl);
   
         /* 2nd arg is TRUE because this count is assoc. with named char */
         putFLpItemName(listElt,fimake((char) charnum,TRUE));
         putFLpOccCount(listElt,fcmake(freq));
         
         pfl = listElt;
        }
     }  
  }  while (beforeEOF);
   
  *status = TRUE ;     /* routine succeeded             */
  return pfl;          /* freq list stored here by now  */
}
