/* 
 *  Obtain next line from input stream, storing in fixed length buffer
 *
 *  inputs:
 *    getcIn  -- a pointer to a function returning a stream of chars
 *               that takes two arguments:
 *                        c -      address of next char to return from stream
 *                        status - address of a status flag
 *                     and returns TRUE iff accessing stream produced a char
 *    linebuf -- buffer to fill with next line
 *    buflen  -- max no. chars linebuf will hold (less 1 for closing '\0')
 *    streamname -- name of current input stream, for error messages;
 *                  if NULL, streamname parameter ignored
 *    linenum -- current line number, 1-indexed, for error messages;
 *               if 0, line number parameter ignored
 *
 *  output values:   
 *     returns TRUE if line recovered
 *     returns FALSE if EOF encountered
 *
 *  side effects:    
 *     linebuf contains next line from stream, on return
 *     truncation warning printed to stderr if buffer length exceeded
 *     routine exits program if abnormal stream condition encountered
 */

#include <stdio.h>
#include "basetypes.h"

   static void streamErr (char *streamname) {
       fprintf(stderr, "?? getline: access error ");
       if (streamname != NULL)  fprintf(stderr," in %s stream",streamname);
       fprintf(stderr, "\n");
       exit(-1);
   }
   static void lenWarn (char *streamname, unsigned lineno, unsigned buflen) {
       fprintf(stderr, "?? getline: line ");
       if (lineno != 0)  fprintf(stderr," %d ",lineno);
       fprintf(stderr, "too long ");
       if (streamname != NULL)  fprintf(stderr," in %s stream",streamname);
       fprintf(stderr, "\n");
       fprintf(stderr, "?? truncated at %d chars\n",buflen);
       fprintf(stderr, "\n");
   }

BOOLEAN getline 
  (BOOLEAN (* getcIn) ( char *c, BOOLEAN *status ),
   char *linebuf, unsigned buflen, char *streamname, unsigned linenum)
{
   unsigned i = 0;              /* current position in line buffer         */
   BOOLEAN beforeEOF = TRUE;             /*  set to true if well runs dry  */
   BOOLEAN atEOL = FALSE;                /*  set to true if well runs dry  */
   BOOLEAN bufOverflow = FALSE;    /* set to true if line too long for buf */

  /* loop:  get one more char from this line until line break, EOF, or err */
  while (beforeEOF && (! atEOL))  {
       char c;  
       BOOLEAN status, instreamStatus;  

       instreamStatus = (*getcIn) (&c, &status);

       if (! instreamStatus)  {
           if (! status)  streamErr(streamname);   /* fatal  */
           else           beforeEOF = FALSE;       /* at end-of-file  */

       } else {
           if (c == '\n')  atEOL = TRUE;           /* break off loop  */
           else {

               if (! bufOverflow) {       /*  accept iff buf not full  */
                   /*   space left in buffer to store next char?  */
                   if (i == buflen) {      /*  no:  warn and flag problem */
                      lenWarn (streamname, linenum, buflen);
                      bufOverflow = TRUE;
                   } else {
                      linebuf[i] = c;
                      i++;
                   }
               }
          }
       } 
  } 

  linebuf[i] = '\0';      /* null-terminate the input line  */
  return beforeEOF;
}
