/*
 *  create a priority queue of one-item Huffman code trees,
 *    ordered by counts,
 *      from list of character frequency counts in file 
 *
 *  Arguments:
 *      getcIn -- a pointer to a function returning a stream of 
 *                incoming chars that takes two arguments:
 *                     c -      address of next char to return from stream
 *                     status - address of a status flag
 *                  and returns TRUE iff accessing stream produced a char
 *
 *             each line in stream consists of two values:
 *               1st in pair:  name of character, in octal (range: 0-277)
 *               2nd in pair:  frequency count, a positive unsigned value
 *                  [initial frequency counts of zero disallowed, due to
 *                   assumptions in dynamic tree-balancing algorithm]
 *
 *          preconditions on stream:  
 *             stream should have unique count for each character
 *             (precondition not audited by routine)
 *
 *      status - pointer to a BOOLEAN value
 *
 *  Outputs:
 *      if init fails, status is set to FALSE on return
 *      if init succeeds,
 *         status is set to TRUE on return
 *         routine returns a pointer to a priority queue of one-item 
 *            Huffman code trees,
 *               ordered by counts,
 *               prioritized by specified character frequency 
 *         priority queue will also have leaf element for the 
 *          "not yet encountered" character, with a count of zero.
 */

#include <stdio.h>            /* standard stream i/o package  */

#include "fitypes.h"
#include "fctypes.h"
#include "fltypes.h"
#include "pqtypes.h"
#include "hitypes.h"

#include "fcfns.h"
#include "fifns.h"
#include "htnfns.h"
#include "htfns.h"
#include "flfns.h"
#include "pqfns.h"
#include "hifns.h"

/*
 *  auxiliary subroutine creates a new one-node huffman tree 
 *     from its item name and frequency count parameters,
 *  then enters that tree into the huffman tree priority queu
 *     named by its first argument,
 *  returning the operation's status in status arg.
 */

static void enter_node_in_huffQ 
    (pPriQ huffmanQ, pFItemname pitem, pFCount pfct, BOOLEAN *status)
{
    pHTreeNode   huffmanTNode;
    pHTree       huffmanT;

    /* first, make the tree */
    huffmanT = htmake(status);
    if (! (*status)) return;

    /* then, make the node */
    huffmanTNode = htnmake(pitem,pfct,status);
    if (! (*status)) {
          /* destroy huff tree--now unneeded */
          htdestroy(huffmanT);  
    }

    /* then, put node at root of tree */
    putHTRoot(huffmanT,huffmanTNode); 

    /* finally, drop the huffman tree into the queue */
    if (! (*status = pqadd(huffmanQ,huffmanT)) ) { 
          /* destroy huff tree--now unneeded */
          htdestroy(huffmanT);  
    }
}

/*** main routine proper ****/

pPriQ hiqueue_init( 
    BOOLEAN (* getcIn) (char *c, BOOLEAN *status), 
    BOOLEAN *status)
{ 
  pPriQ        huffmanQ;
  pFList       freqlist, flElt;

  /*
   * convert character count frquencies from input file to list form.
   * if conversion failed, give up:  can't complete task.
   */
  freqlist = flinit(getcIn, status);
  if (! (*status)) {
     /* set up result -- status initialized to FALSE */
     return nilPriQ;
  }
  
  /*
   *  allocate priority queue; allow one slot for each item in list of
   *    character counts, plus one slot for the nil character count
   *  if allocation failed, give up:  can't complete task.
   */
  huffmanQ = pqmake(fllength(freqlist)+1, status);
  if (! (* status)) {
     fldestroy(freqlist);     /* destroy frequency list--now unneeded */
     return nilPriQ;          /* result shows false */
  }

  /*
   *  make and drop a one-node tree that contains the "empty huffman count"
   *  into the queue--needed for dynamic Huffman processing.
   *    -.  item will be unnamed
   *    -.  frequency count will be zero
   *  if allocation required by add failed, give up:  can't complete task
   */
  enter_node_in_huffQ (huffmanQ,fimake(' ', FALSE),fcmake(0),status);
  if (! (*status)) {
     /* destroy frequency list, Huffman queue--now unneeded */
      fldestroy(freqlist);  
      pqdestroy(huffmanQ);  
      return nilPriQ;
  }

  /*
   *  now convert each of the items 
   *     in the frequency count list into a Huffman code tree,
   *  and drop each Huffman code tree into the priority queue
   */
  flElt = freqlist;
  while (! isNilFL(flElt)) {

       enter_node_in_huffQ 
          (huffmanQ, getFLpItemName(flElt), getFLpOccCount(flElt), status);

       if (! (*status)) {
          /* destroy frequency list, Huffman queue--now unneeded */
          fldestroy(freqlist);  
          pqdestroy(huffmanQ);  
          return nilPriQ;
       }

      /* and step to next element in the freqUency list */
      flElt = getFLNext(flElt);
  }

  /*   clean up the intermediate storage, and go home */

  fldestroy(freqlist);  /* destroy frequency list--now unneeded */
  *status = TRUE;
  return huffmanQ;
}
