/*
 *  master function in module, httreedump.c, 
 *     displays a Huffman code tree to a maximum convenient depth 
 *
 *  two arguments:
 *      first argument:    tree to display
 *      second argument:   line width for display
 *
 *  naive algorithm used to display tree
 *  -.  tree spread across screen in manner similar to full binary tree
 *  -.  tree displayed one level at a time
 *  -.  worklist used to track printing of tree
 *      -.  worklist, when displaying level n, has 2^n elements
 *              (note: first level is level 0)
 *      -.  worklist elts ordered from left to right in worklist
 *              by order of appearance in inorder traversal
 *      -.  nil worklist elements serve as placeholders for
 *              "spaces beyond leaves" in original tree
 *
 */
    
#include <stdio.h>

#include "httypes.h"

#include "htnfns.h"
#include "wlfns.h"
#include "htfns.h"

    /*
     ******************************************************************
     *  convert pwl, a ptr to worklist of elts at current level of tree
     *          to ptr to worklist of elts at next deeper level of tree
     *
     *  return TRUE if any items in tree's next level
     *****************************************************************
     */

   static BOOLEAN next_generation_worklist( pWList pwl )
   {
       pWList cur_rec = pwl;

       /*
        * set next variable TRUE during loop evaluation if any subtrees 
        * of trees currently in worklist are nonempty -- i.e., if more 
        * to print at next level down
        */

       BOOLEAN next_level_has_nodes = FALSE;
		
       while (! isNilWL(cur_rec)) {
           pWList new_rec;

           /*
            *  work list for next level down will have two elements 
            *  in place of element at current level:  
            *     update cur_rec to corresp. to its left sibling
            *     make new_rec corresp. to right sibling for current
            */

           if (isNilWLTreeNode(cur_rec)) {

              /*
               *  nothing to print in subtree at next level. 
               *  (note: cur_rec already points to nil tree)  
               */
              new_rec = wlmake(nilHTNode);

            } else {
              pHTreeNode leftChild  = getHTNLeft(getWLTreeNode(cur_rec));
              pHTreeNode rightChild = getHTNRite(getWLTreeNode(cur_rec));

               /*
                *  this node may have non-nil children
                *   put left, right children in cur_rec, new_rec
                *   check and see if any nodes at next level yet found
                */

	       next_level_has_nodes = 
                    next_level_has_nodes 
                        || (! isNilHTNode(leftChild))
	                   || (! isNilHTNode(rightChild));

               new_rec = wlmake(rightChild);
               putWLTreeNode(cur_rec,leftChild);
            }
            /*  tie the new record into the list */

	    putWLNext(new_rec, getWLNext(cur_rec));
	    putWLNext(cur_rec, new_rec);

            cur_rec = getWLNext(new_rec);
        }
        return next_level_has_nodes;
    }


/*
 * *************************************************
 *
 *  finally, the master function itself
 *
 * *************************************************
 */

void httreedump (pHTree pt, unsigned disp_line_length)
{
   if (isNilHT(pt)) {
       printf("\n\n tree is empty \n\n");
       return;
   } 

  {
  /*  set up worklist to start printing at tree root  */
   pWList worklist = wlmake(getHTRoot(pt)); 
   pWList cur_rec;
   unsigned level = 0;

   /*
    *  iteratively print the tree, level by level, until the next level
    *  contains only nil items.  note: the "next_generation_worklist"
    *  routine will always grow the number of items in the worklist  
    *  by a power of 2.
    */

   do  { 
        unsigned num_items_on_level = 1 << level;
        unsigned spaces_for_each_item = disp_line_length / num_items_on_level;
        unsigned leftover_spaces = disp_line_length % num_items_on_level;

        /*
         *  partition the spaces on the current line according to level
         *  (i.e., according to number of items to print on this level
         *
         */
        cur_rec = worklist;

        /*
         * repeatedly display nodes on current level by scanning worklist.
         * print nil nodes as spaces;  dump non-nil nodes using htnodedump()
         */ 
        while (! isNilWL(cur_rec))  {
            unsigned spaces_for_this_item = spaces_for_each_item;

            if (leftover_spaces > 0) {
                spaces_for_this_item++;
                leftover_spaces--;
            }

            if (! isNilWLTreeNode(cur_rec)) {
               if (! htnodedump(getWLTreeNode(cur_rec),spaces_for_this_item)) {
                  /*  
                   * can't print any more levels of tree if element-displaying
                   * logic can't pack items into current field width
                   */
                  goto screen_too_narrow_exception;
               }
            } else {
               printf("%*c", spaces_for_this_item, ' ');
            }

            cur_rec = getWLNext(cur_rec);
        }
        printf("\n");
        level++;

   } while (next_generation_worklist (worklist));
     /* generate next level of display, if there's more to show  */

   /* done with display:  clean up, go home.  */
   printf("\n");
   wldestroy(worklist);
   return;

/* line too narrow:  can't fit any more items from tree into display  */
screen_too_narrow_exception:
   printf("\n . . . . \n ");
   wldestroy(worklist);
 }
}
