/*
 *  Program builds a (static) Huffman code tree from a file of
 *  <character, frequency count> pairs
 *
 *  Command line arguments:
 *     -ffilename:  name of initial counts file
 *         each line in file consists of two values:
 *            1st in pair:  name of character, in octal (range: 0-277)
 *            2nd in pair:  frequency count, a positive unsigned value
 *                [initial frequency counts of zero disallowed, due to
 *                 assumptions in dynamic tree-balancing algorithm]
 *
 *  Preconditions on file:  
 *      -.  initial counts file should have unique count for each character
 *             (precondition not audited by program)
 *      -.  initial counts file line lengths should not exceed 80 characters
 *             (violation of condition flagged by program; excess chars ignored)
 *
 *  Effect:
 *     -.  computes Huffman code tree for file,
 *             with special leaf element for "characters not yet seen"
 *     -.  allows user to invoke tree routine that displays resulting tree
 *
 *
 *  Options:
 *
 */

#include <stdio.h>

#include "pqtypes.h"

#include "htfns.h"
#include "pqfns.h"
#include "hifns.h"

void usage(char *namestring) {
  fprintf(stderr,
  "?? usage: %s -ffilename\n",namestring);
  fprintf(stderr,
   "   -f:   name of file containing counts data\n");
};

/* 
 *  treat frequency input stream as global, to allow it to be 
 *        wrapped in character-by-character access functions;
 *  needed to allow input processing to proceed separately.
 *
 */
FILE *fpfreq;    

int getc_fpfreq ( char *c, BOOLEAN *s )
{ int temp = getc(fpfreq);
  *s = ( temp == EOF ) || ( temp >= 0 );   // did we get something meaningful--char or EOF?
  *c = (char) temp;                        // value returned, as char
  return ( temp >= 0 );                    // did we get a char?
}

main(int argc, char *argv[])
{
   BOOLEAN file_named = FALSE;
   char *pgmname = *argv;
   pHTree huffmanTree;

  /*
   * ---- check command line arguments  -----
   */

  while ((--argc > 0) && ((*++argv)[0] == '-')) {
     char *option = argv[0]+1;

     switch (*option++) {
        case 'f':
            fpfreq = fopen(option,"r"); 
            if (fpfreq == NULL) {
               fprintf(stderr,"?? could not open input file\n");
               fprintf(stderr,"file name: %s\n\n",option);
               exit(1);
            }
            file_named = TRUE;   /* input file specified */
            break;

	    default:  
	         fprintf(stderr,"?? unrecognized option: %c\n\n",*--option);
	         usage(pgmname);
	         exit(1);
	  }
   }
   if ((argc != 0) || (! file_named)) {
       usage(pgmname); 
       exit(1);
   }

   /*
    * first, build priority queue of one-count huffman code trees 
    * from counts data in input file.
    * then, build master static Huffman tree from worklist.
    */
   
   {BOOLEAN status;
    pPriQ work_queue;

    work_queue = hiqueue_init(getc_fpfreq, &status);
    if (! status) exit(-1);

    huffmanTree = hitree_build(work_queue, &status);
    if (! status) exit(-1);
   }

   /* 
    *  now, display resulting tree interactively
    */
    
   htshowtree(huffmanTree);

   exit(0);    /* program concluded successfully */
}
