/*
 *  combine a priority queue of one-element Huffman trees
 *    into a finished Huffman tree
 *
 *  Arguments:
 *    huffmanQ -  the queue of Huffman trees
 *      precondition:  priority queue must have at least one element
 *
 *    status - a pointer to a value of type BOOLEAN
 *
 *  Outputs:
 *    if tree build fails, status flag is set to FALSE
 *    if tree build succeeds, 
 *       status flag is set to TRUE 
 *       function returns a Huffman code tree 
 *          built from trees in initial priority queue, with
 *          intermediate nodes set to appropriate weights
 *
 *  Side effects:
 *    Contents of priority queue consumed, and queue deallocated
 */

#include <stdio.h>            /* standard stream i/o package  */

#include "fltypes.h"
#include "pqtypes.h"

#include "fcfns.h"
#include "fifns.h"
#include "htfns.h"
#include "pqfns.h"

#include "static_tree_build.h"

pHTree static_tree_build( pPriQ huffmanQ, BOOLEAN *status )
{ 
  if (isEmptyPriQ(huffmanQ)) {
     fprintf(stderr,"?? error in static_tree_build");
     fprintf(stderr,"?? initial Huffman queue is empty");

     /*  destroy the queue and quit */
     pqdestroy(huffmanQ);

     *status = FALSE;
     return nilHT;
  }

  /* until there's just one tree in the huffman queue,
   *  -.  remove the first two trees
   *  -.  combine them into a new Huffman tree
   *  -.  put the new Huffman tree back into the priority queue
   */

  /*  returns used to exit this loop  */
  while (TRUE) {
      pHTree           firstT, secondT, newT;
      pHTreeNode       huffmanTNode;

      /* get first code tree from priority queue  */
      firstT = getPQElt(getPriQelt(huffmanQ, 1));    
      if (! pqrem(huffmanQ)) {
          fprintf(stderr,"?? programmer error in queue_init");
          fprintf(stderr,"?? couldn't remove item from nonempty queue");
          exit(-1);
      }

      /* if no more code trees, we're done coalescing items from queue  */
      if (isEmptyPriQ(huffmanQ)) {

          /*  destroy the queue -- no longer needed */
          pqdestroy(huffmanQ);

          *status = TRUE;
          return firstT;
      }

      /* get second code tree from priority queue  */
      secondT = getPQElt(getPriQelt(huffmanQ, 1));   
      if (! pqrem(huffmanQ)) {
          fprintf(stderr,"?? programmer error in queue_init");
          fprintf(stderr,"?? couldn't remove item from nonempty queue");
          exit(-1);
      }
   
      /*
       * cons first, second code trees to form new code tree.
       *  -.  root node is nameless
       *  -.  frequency count combines counts of two child trees
       */
      newT = htjoin(firstT,secondT,status);
      if (! status) {        /* the cons failed?  allocation error? */
          /*  destroy queue and left, right trees -- no longer needed */
          pqdestroy(huffmanQ);
          htdestroy(firstT);
          htdestroy(secondT);

          *status = FALSE;
          return nilHT;
      }

      /*
       * put new tree back into queue for continued coalescing
       */
      if (! pqadd(huffmanQ,newT)) {
          /*  destroy tree and queue -- no longer needed */
          pqdestroy(huffmanQ);
          htdestroy(newT); 

          *status = FALSE;
          return nilHT;
      }
  }
}
